# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from enum import Enum
from azure.core import CaseInsensitiveEnumMeta


class GeoReplicationStatusType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The status of the secondary location."""

    LIVE = "live"
    BOOTSTRAP = "bootstrap"
    UNAVAILABLE = "unavailable"


class OdataMetadataFormat(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """OdataMetadataFormat."""

    APPLICATION_JSON_ODATA_NOMETADATA = "application/json;odata=nometadata"
    APPLICATION_JSON_ODATA_MINIMALMETADATA = "application/json;odata=minimalmetadata"
    APPLICATION_JSON_ODATA_FULLMETADATA = "application/json;odata=fullmetadata"


class ResponseFormat(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """ResponseFormat."""

    RETURN_NO_CONTENT = "return-no-content"
    RETURN_CONTENT = "return-content"
